(*  Title:      Pretty_Print_Polynomials.thy
    Author:     Patrick Riegler
*)



theory Pretty_Print_Polynomials
imports Main


begin

section \<open> Distributive Polynomials \<close>
ML \<open>
datatype comp = NDEF | Is_Smaller of bool
type order = (int list * int list -> comp)


type monom = (int * int) list  (* sparse monoms *)

 
datatype 'a sparse_polynom = SPoly of ('a * monom) list
datatype 'a dense_polynom = DPoly of 'a list * order
\<close>


text \<open> auxiliary functions \<close>
ML \<open>

(** <lex **)
fun lex_order (nil, nil) : comp = Is_Smaller(false)
  | lex_order ([h1] : int list, [h2] : int list) : comp =
        if (h1 < h2) then Is_Smaller(true)
        else Is_Smaller(false)
  | lex_order (h1::l1 : int list, h2::l2 : int list) : comp =
        if (h1 < h2) then Is_Smaller(true)
        else if (h1 > h2) then Is_Smaller(false)
        else lex_order (l1, l2)
  | lex_order (_, _) = error "lex_order: invalid arguments"


fun string_of_comp (Is_Smaller(b) : comp) : string = Bool.toString b
  | string_of_comp (NDEF : comp) : string = "Not comparable"


fun contains (_, nil) : bool = false
  | contains (e, h::t) : bool = if (e = h) then true else contains (e, t)





fun partition (_, nil) = (nil, nil)
 |  partition (pivot, h::t) =
        let
            val (smaller, greater) = partition(pivot, t)
        in
            if h < pivot then (h::smaller, greater)
            else (smaller, h::greater)
        end;


fun quicksort nil = nil
 |  quicksort [h] = [h]
 |  quicksort (h::t) =
        let
            val (smaller, greater) = partition(h, t)
        in
            quicksort(smaller) @ [h] @ quicksort(greater)
        end;




(** quicksort int lists **)
fun list_partition (_, nil, _) = (nil, nil)
 |  list_partition (pivot, h::t, f : order) =
        let
            val (smaller, greater) = list_partition(pivot, t, f)
        in
            if f(h, pivot) = Is_Smaller(true) then (h::smaller, greater)
            else (smaller, h::greater)
        end;

fun list_qsort (nil, _) = nil
 |  list_qsort ([h], _) = [h]
 |  list_qsort (h::t, f : order) =
        let
            val (smaller, greater) = list_partition(h, t, f)
        in
            list_qsort(smaller, f) @ [h] @ list_qsort(greater, f)
        end;


fun cond_append (l, e, n) = if (length (l) < n) then l@[e] else l


fun is_nat_list nil = false
  | is_nat_list [h] = h >= 0
  | is_nat_list (h::t) = 
      if h >= 0 then
         is_nat_list t
      else
         false


fun check_vars vars = 
      if is_nat_list (vars) then
        quicksort (vars)
      else
        error "check_vars: Invalid variables!"


fun generate_var_names (nil, e, k, n) = 
      if (k < n) then 
        generate_var_names ([e ^ "_" ^ string_of_int (k)], e, k+1, n) 
      else 
        nil
  | generate_var_names (l, e, k, n) = 
      if (k < n) then 
        generate_var_names (l@[e ^ "_" ^ string_of_int (k)], e, k+1, n) 
      else 
        l

\<close>



text \<open> Monom functions \<close>
ML \<open>

fun partition_monom ([(_,_)] : monom, nil) = (nil, nil)
 |  partition_monom ([(v,e)] : monom, (u,e')::t : monom) =
        let
            val (smaller, greater) = partition_monom([(v,e)], t)
        in
            if u < v then ((u,e')::smaller, greater)
            else (smaller, (u,e')::greater)
        end
  | partition_monom (_, _) = error "partition_monom: invalid argument" 

fun qsort_monom (nil) = nil
 |  qsort_monom ([(v,e)] : monom) = [(v,e)]
 |  qsort_monom ((v,e)::t : monom) =
        let
            val (smaller, greater) = partition_monom([(v,e)], t)
        in
            qsort_monom(smaller) @ [(v,e)] @ qsort_monom(greater)
        end;

fun arrange_monoms nil = nil
  | arrange_monoms ( [(c,m)] : ('a * monom) list) = [(c,qsort_monom (m))]
  | arrange_monoms ( (c,m)::t : ('a * monom) list) =  (c,qsort_monom (m))::arrange_monoms (t)


(** exponent as list**)
fun exponent ([(_,e)] : monom) = [e]
  | exponent ((_,e)::t : monom) = e::exponent(t)
  | exponent (_ : monom) = error "exponent: invalid argument"

(** does not count [v, 0] **)
fun vars_of_monom (nil : monom, vars : int list) = vars
 |  vars_of_monom ([(v,_)] : monom, vars : int list) =
        if contains(v,vars) then vars else vars@[v]
 |  vars_of_monom ((_,0)::m : monom, vars : int list) =
        vars_of_monom(m, vars)
 |  vars_of_monom ((v,_)::m : monom, vars : int list) =
        if contains(v, vars) then vars_of_monom(m, vars) else vars_of_monom(m,vars@[v])

fun var_to_name (v ,[h] : int list, [n] : string list) = 
      if (v = h) then n else error "var_to_name: variable not in list"
  | var_to_name (v, h::t : int list, n::l : string list) =
       if (v = h) then n else var_to_name(v, t, l)
  | var_to_name (_, nil, _) = error "var_to_name: variable list is nil"
  | var_to_name (_, _, nil) = error "var_to_name: name list is nil"


fun sdist_substitute_names ([(_,_)] : monom, nil, _) = ""
  | sdist_substitute_names ([(_,0)] : monom, _ : int list, _ : string list) = ""
  | sdist_substitute_names ([(v,1)] : monom, vars : int list, names : string list) =
        var_to_name (v, vars, names)
  | sdist_substitute_names ([(v,e)] : monom, vars : int list, names : string list) =
        var_to_name (v, vars, names) ^ "^" ^ string_of_int(e)
  | sdist_substitute_names (_, _, _) = error "sdist_substitute_names: invalid arguments"


fun monom_with_names (nil : monom, _, _) = ""
 |  monom_with_names ([(v,e)]: monom, vars : int list, var_names : string list) =
        sdist_substitute_names([(v,e)], vars, var_names)
 |  monom_with_names ((v,e)::m : monom, vars : int list, var_names : string list) =
        sdist_substitute_names([(v,e)], vars, var_names)  ^ " " ^ monom_with_names(m, vars, var_names)


fun string_of_monom (m : monom, varNames : string list) =
let
    val vars = vars_of_monom (m, nil)
in
    if (length vars <= length varNames) then
        monom_with_names (m, vars, varNames) ^ "\n"
    else error "string_of_monom: Not enough variables provided"
end;

\<close>




text \<open> Sparse Polynom functions \<close>
ML \<open>

(** partition spolynom by power products **)
fun partition_spoly_by_pp ([(_,_)] : ('a * monom) list, _) = (nil, nil)
  | partition_spoly_by_pp ((c,m)::(c',m')::p : ('a * monom) list, f : order) =
        let
            val (smaller, greater) = partition_spoly_by_pp((c,m)::p, f)
        in
            if f(exponent m', exponent m) = Is_Smaller(true) then ((c',m')::smaller, greater)
            else (smaller, (c',m')::greater)
        end
  | partition_spoly_by_pp (_,_) = error "partition_spoly_by_pp: invalid argument"

(** quicksort spolynom by power products **)
fun qsort_spoly_by_pp (nil, _) = nil
  | qsort_spoly_by_pp ([(c,m)] : ('a * monom) list, _) = [(c,m)]
  | qsort_spoly_by_pp ((c, m)::p : ('a * monom) list, f : order) =
        let
            val (smaller, greater) = partition_spoly_by_pp((c, m)::p, f)
        in
            qsort_spoly_by_pp(smaller, f) @ [(c,m)] @ qsort_spoly_by_pp(greater, f)
        end

fun sort_spoly (SPoly l, f : order) = SPoly (qsort_spoly_by_pp (arrange_monoms(l), f))


fun vars_of_spoly (SPoly [(_, nil)] : 'a sparse_polynom, vars : int list) : int list = vars
  | vars_of_spoly (SPoly [(_,m)] : 'a sparse_polynom, vars : int list) : int list =
        vars_of_monom(m, vars)
  | vars_of_spoly (SPoly ((_,m)::p) : 'a sparse_polynom, vars : int list) : int list =
        vars_of_spoly (SPoly p, vars_of_monom(m, vars))
  | vars_of_spoly (_, _) = error "vars_of_spoly: invalid arguments"



(* We are given 2 lists [a,b,c,...] and [x,y,z,...], we want to define a function which takes 
  ONE argument and maps  a \<rightarrow> x, b \<rightarrow> y, c \<rightarrow> z, ... How can we do that ?
*)


fun spoly_with_names (SPoly [(_, nil)] : 'a sparse_polynom, _, _, _) : string = ""
  | spoly_with_names (SPoly [(c,m)] : 'a sparse_polynom, vars : int list, var_names : string list, toString : 'a -> string) : string =
        toString (c) ^ 
        " " ^ 
        monom_with_names (m, vars, var_names)
  | spoly_with_names (SPoly ((c,m)::p) : 'a sparse_polynom, vars : int list, var_names : string list, toString : 'a -> string) : string =
        toString (c) ^ 
        " " ^ 
        monom_with_names (m, vars, var_names)  ^ 
        " + " 
        ^ spoly_with_names(SPoly p, vars, var_names, toString)
  | spoly_with_names (_, _, _, _) = error "spoly_with_names: invalid arguments"


fun string_of_spoly (p : 'a sparse_polynom, varNames : string list) =
let
    val vars = check_vars (vars_of_spoly (p, nil))
in
    if (length vars <= length varNames) then
        spoly_with_names (p, vars, varNames,@{make_string}) ^ "\n"
    else
         spoly_with_names (p, vars, generate_var_names (nil, hd(varNames), 0, length(vars)), @{make_string}) ^"\n"
end

\<close>




text \<open> Dense Distributive Polynom functions (How to implement order in multivariate case ? list ?) \<close>
ML \<open>

\<close>





section \<open> Recursive Polynomials \<close>


ML \<open>

datatype 'a srec_monom = SRec_Monom_base of 'a * (int * int) | SRec_Monom of 'a srec_monom * (int * int)   (* sparse recursive monoms *)

 
datatype 'a srec_polynom = SRec_Poly of 'a srec_monom list
\<close>


ML \<open>
fun vars_of_srec_monom (SRec_Monom_base (_, (_,0)) : 'a srec_monom, vars : int list) = vars
  | vars_of_srec_monom (SRec_Monom_base (_, (v,_)) : 'a srec_monom, vars : int list) = 
      if contains(v,vars) then vars else vars@[v]
  | vars_of_srec_monom (SRec_Monom (m, (_,0)) : 'a srec_monom, vars : int list) = 
      vars_of_srec_monom(m, vars)
  | vars_of_srec_monom (SRec_Monom (m, (v,_)) : 'a srec_monom, vars : int list) = 
      if contains(v, vars) then vars_of_srec_monom(m, vars) else vars_of_srec_monom(m,vars@[v])


(** uses variables in wrong order **)
fun srec_monom_with_names (SRec_Monom_base (c, (_,0)), _, _, toString) = toString c
  | srec_monom_with_names (SRec_Monom_base (c, (v,1)), vars, names, toString) = 
        toString (c) ^ " " ^ var_to_name (v, vars, names)
  | srec_monom_with_names (SRec_Monom_base (c, (v,e)), vars, names, toString) = 
        toString (c) ^ " " ^ var_to_name (v, vars, names) ^ "^" ^ string_of_int(e)
  | srec_monom_with_names (SRec_Monom (m, (_,0)), vars, names, toString) =
        srec_monom_with_names (m, vars, names, toString)
  | srec_monom_with_names (SRec_Monom (m, (v,1)), vars, names, toString) =
        srec_monom_with_names (m, vars, names, toString) ^ " " ^ var_to_name (v, vars, names)
  | srec_monom_with_names (SRec_Monom (m, (v,e)), vars, names, toString) =
        srec_monom_with_names (m, vars, names, toString) ^ " " ^ var_to_name (v, vars, names) ^ "^" ^ string_of_int(e)


fun string_of_srec_monom (m, varNames) =
let
    val vars = vars_of_srec_monom (m, nil)
in
    if (length vars <= length varNames) then
        srec_monom_with_names (m, vars, varNames, @{make_string})
    else error "string_of_srec_monom: Not enough variables provided"
end;






fun vars_of_srec_poly (SRec_Poly nil, vars) = vars
  | vars_of_srec_poly (SRec_Poly [m], vars) =  vars_of_srec_monom(m, vars)
  | vars_of_srec_poly (SRec_Poly (m::p), vars) =
      vars_of_srec_poly (SRec_Poly p, vars_of_srec_monom(m, vars))


fun srec_poly_with_names (SRec_Poly nil, _, _, _) = ""
  | srec_poly_with_names (SRec_Poly [m], vars, var_names, toString) =
      srec_monom_with_names (m, vars, var_names, toString)
  | srec_poly_with_names (SRec_Poly (m::p), vars, var_names, toString) =
      srec_monom_with_names (m, vars, var_names, toString)  ^ 
      " + " 
      ^ srec_poly_with_names(SRec_Poly p, vars, var_names, toString)


fun string_of_srec_poly (p, varNames) =
let
   val vars = check_vars (vars_of_srec_poly (p, nil))
in
  if (length vars <= length varNames) then
    srec_poly_with_names (p, vars, varNames,@{make_string}) ^ "\n"
  else
    srec_poly_with_names (p, vars, generate_var_names (nil, hd(varNames), 0, length(vars)), @{make_string}) ^"\n"
end

\<close>








text  \<open> test \<close>

ML \<open>
val pol : int sparse_polynom = SPoly [(1, [(1,3), (5,4), (3,2)]), (2, [(1,1), (3,5), (2,6)])];
writeln (string_of_spoly(pol, ["X", "Y", "Z", "W"]));

arrange_monoms ([(1, [(2,1), (1,3), (3,2)]), (2, [(4,7), (3,5), (1,1)])]);

val pol' = sort_spoly ( pol, lex_order);

writeln (string_of_spoly(pol', ["X", "Y", "Z", "W"]));

\<close>
ML \<open>
val l = list_qsort ([[2,2,4], [2,2,3], [1,1,1]], lex_order);
writeln (@{make_string} l);
\<close>
ML \<open>
val c = lex_order ([2,2,3], [2,2,3]);
writeln (string_of_comp (c) ^ "\n")
\<close>

ML \<open>
generate_var_names(nil, "X", 0, 3);

writeln (string_of_spoly(pol', ["X", "Y", "Z"]));
\<close>


ML \<open>


val srm : int srec_monom = SRec_Monom( SRec_Monom_base (2, (1,3)), (3,5));

string_of_srec_monom (srm, ["X","Y"]);


val srp : int srec_polynom = SRec_Poly( [SRec_Monom( SRec_Monom_base (2, (1,3)), (3,5)), SRec_Monom( SRec_Monom( SRec_Monom_base (7, (2,4)), (4,1)), (1,9))]);

writeln (string_of_srec_poly(srp, ["X", "Y", "Z", "W"]))

\<close>

end