(* see 
  https://github.com/larsrh/leon/blob/8a01cfea513376821c02a2390b3a7ed3266d336c/
    src/main/isabelle/Isabelle_Leon/Protocol.thy
  Author: Lars Hupel
*)

theory Protocol_Leon
imports Main
begin
ML_file "~~/src/HOL/TPTP/sledgehammer_tactics.ML"

ML{*
local
structure Assertions = Named_Thms(
  val name = @{binding assertion}
  val description = "Leon Assertions"
)
structure Trace = Proof_Data(
  type T = int list
  fun init _ = []
)
val session_lthy = 
  Synchronized.var "isabelle_leon.session_lthy" (Inttab.empty: local_theory Inttab.table)

fun guard_initialized (id, req) f =
  let
    fun upd table =
      case Inttab.lookup table id of
        NONE => (NONE, table)
      | SOME lthy =>
          let
            val trace = Trace.get lthy
            val (res, lthy) = f lthy
            val lthy = Trace.put (req :: trace) lthy
          in
            (SOME res, Inttab.update (id, lthy) table)
          end
  in
    case Synchronized.change_result session_lthy upd of
      NONE => error "Not initialized"
    | SOME res => res
  end

fun all_frees (Free (name, typ)) = [(name, typ)]
  | all_frees (t $ u) = union (op =) (all_frees t) (all_frees u)
  | all_frees (Abs (_, _, t)) = all_frees t
  | all_frees _ = []

val dummify_types = map_types (K dummyT)

fun register_terms ts lthy =
  let
    val ts' = Syntax.check_terms lthy ts |> map HOLogic.mk_Trueprop
    val frees = distinct (op =) (maps all_frees ts') 
      |> filter (not o Variable.is_declared lthy o fst)
    val (free_names', lthy) = Variable.add_fixes (map fst frees) lthy
    val frees' = map Free (free_names' ~~ map snd frees)
    val frees = map Free frees
    val ts'' = map (subst_free (frees ~~ frees')) ts'
    val _ = Pretty.block (map (Syntax.pretty_term lthy) ts'') |> Pretty.writeln
  in
    (ts'', fold Variable.declare_term ts'' lthy)
  end

in

structure Operations = struct

fun init id thy =
  let
    fun upd table =
      case Inttab.lookup table id of
        NONE => Inttab.update_new (id, Named_Target.init "" thy) table
      | SOME _ => error "Local theory already initialized"
  in
    Synchronized.change session_lthy upd;
    XML.Encode.string (Context.theory_name thy)
  end

fun mk_numeral raw_n raw_typ =
  let
    val n = XML.Decode.int raw_n
    val typ = Type (XML.Decode.string raw_typ, [])
    fun wrap numeral = HOLogic.numeral_const typ $ numeral
    val numeral =
      if n = 0 then
       Const (@{const_name "zero_class.zero"}, typ)
      else if n > 0 then
        wrap (HOLogic.mk_numeral n)
      else
        Const (@{const_name "uminus"}, dummyT) $ wrap (HOLogic.mk_numeral (~n))
  in
    Term_XML.Encode.term numeral
  end;

fun mk_fresh id raw_ns =
  let
    val ns = map XML.Decode.string raw_ns
    val f = Variable.variant_fixes ns
  in
    XML.Encode.list XML.Encode.string (guard_initialized id f)
  end

fun def_data id raw_data =
  let
  val decoder =
    XML.Decode.pair
      (XML.Decode.pair
        XML.Decode.string
        (XML.Decode.list Term_XML.Decode.typ))
      (XML.Decode.list
        (XML.Decode.pair
          XML.Decode.string
          (XML.Decode.list
            (XML.Decode.pair XML.Decode.string Term_XML.Decode.typ))))
  val encoder =
    XML.Encode.list
      (XML.Encode.list
        (XML.Encode.pair
          Term_XML.Encode.term
          (XML.Encode.list Term_XML.Encode.term)))
  fun transform ((name, tparams), cases) =
    let
      val all_names = map fst cases @ map fst (maps snd cases)
      val nctxt = Name.make_context all_names
      val cases' = Name.invent_names nctxt "is" cases
      fun transform_tparam tp = (NONE, (tp, @{sort type}))
      fun transform_field (sel, typ) = (Binding.name sel, typ)
      fun transform_case (disc, (n, fields)) =
        (((Binding.name disc, Binding.name n), map transform_field fields), NoSyn)
      val head = ((map transform_tparam tparams, Binding.name name), NoSyn)
    in
      (((head, map transform_case cases'), (Binding.empty, Binding.empty)), [])
    end
  val decoded = map decoder raw_data
  val data = map transform decoded
  fun f lthy =
    let
      val thy_name = Context.theory_name (Proof_Context.theory_of lthy)
      val spec = ((false, false), data)
      val lthy = BNF_FP_Def_Sugar.co_datatypes BNF_Util.Least_FP BNF_LFP.construct_lfp spec lthy
      fun get_sugar ((name, _), _) =
        let
        val {discs, selss, ...} =
          the (Ctr_Sugar.ctr_sugar_of lthy (Long_Name.qualify thy_name name))
        in
          map dummify_types discs ~~ map (map dummify_types) selss
        end
      val sugars = map get_sugar decoded
    in
      (sugars, lthy)
    end
  in
  encoder (guard_initialized id f)
  end

fun assert id raw_ts =
  let
    val ts = map Term_XML.Decode.term raw_ts
    fun f lthy =
      let
        val (ts', lthy) = register_terms ts lthy
        val cts = map (cterm_of (Proof_Context.theory_of lthy)) ts'
        val (thms, lthy) = Assumption.add_assumes cts lthy
        val lthy = fold (Context.proof_map o Assertions.add_thm) thms lthy
      in
        ((), lthy)
      end
  in
    (* FIXME return value TBD *)
    XML.Encode.unit (guard_initialized id f)
  end

fun contradiction id =
  let
    fun f lthy =
      let
        val prop = @{term "Trueprop False"}
        val assms = Assertions.get lthy
        val sledgehammer_opts = [("provers", ATP_Proof.spassN ^ " " ^ ATP_Proof.eN), ("timeout", "3")]
        val sledgehammer_thms = {add = [], del = [], only = false}
        fun tac ctxt =
          Method.insert_tac assms THEN'
          (SELECT_GOAL (TRY (auto_tac ctxt))) THEN_ALL_NEW
            FIRST'
              [Cooper.tac false [] [] ctxt,
              Sledgehammer_Tactics.sledgehammer_with_metis_tac ctxt sledgehammer_opts sledgehammer_thms]
        fun meth {context, ...} = tac context 1
      in
        (can (Goal.prove lthy [] [] prop) meth, lthy)
      end
  in
    XML.Encode.bool (guard_initialized id f)
  end

fun trace id =
  XML.Encode.list XML.Encode.int (guard_initialized id (`Trace.get))
val setup = Assertions.setup
end
end
*}

text{* Setup *}
setup{* Operations.setup *}

ML{*
local
fun response sys_id req_id: Properties.T =
  [(Markup.functionN, "isabelle_leon_response"),
   ("sys_id", Markup.print_int sys_id),
   ("req_id", Markup.print_int req_id)];

val encode_result = XML.Encode.variant
  [fn Exn.Res a => ([], a),
   fn Exn.Exn exn => ([], XML.Encode.string (@{make_string} exn))]
in
val _ = Session.protocol_handler "leon.solvers.isa.System$Handler"
val _ = Isabelle_Process.protocol_command "isabelle_leon"
  (fn sys_id :: req_id :: cmd :: args =>
    let
      val sys_id = Markup.parse_int sys_id
      val req_id = Markup.parse_int req_id
      val id = (sys_id, req_id)
      fun exec f x =
        (Future.fork (fn () =>
          let
            val res = Exn.interruptible_capture f x
            val yxml = YXML.string_of_body (encode_result res)
          in
            Output.protocol_message (response sys_id req_id) [yxml]
          end);
        ())
      val args = map YXML.parse_body args
    in
      case (cmd, args) of
        ("init", []) => exec (Operations.init sys_id) @{theory}
      | ("mk_numeral", [n, typ]) => exec (Operations.mk_numeral n) typ
      | ("mk_fresh", ns) => exec (Operations.mk_fresh id) ns
      | ("def_data", data) => exec (Operations.def_data id) data
      | ("assert", ts) => exec (Operations.assert id) ts
      | ("contradiction", []) => exec Operations.contradiction id
      | ("trace", []) => exec Operations.trace id
      | _ => error "isabelle-repl: unknown command"
    end)
end
*}

ML{*
@{binding assertion};
Proof_Context.theory_of;
*}ML{*
Session.protocol_handler : string -> unit;
Isabelle_Process.protocol_command : string -> (string list -> unit) -> unit
*}ML{*
fun response sys_id req_id: Properties.T =
  [(Markup.functionN, "isabelle_leon_response"),
   ("sys_id", Markup.print_int sys_id),
   ("req_id", Markup.print_int req_id)];
response : int -> int -> Properties.T
*}ML{*
val encode_result = XML.Encode.variant
  [fn Exn.Res a => ([], a),
   fn Exn.Exn exn => ([], XML.Encode.string (@{make_string} exn))];
encode_result (*= fn: Encode.body Exn.result XML.Encode.T *)
*}ML{*
fun mk_numeral raw_n raw_typ = (* = fn: Decode.body -> Decode.body -> Encode.body*)
  let
    val n = XML.Decode.int raw_n
    val typ = Type (XML.Decode.string raw_typ, [])
    fun wrap numeral = HOLogic.numeral_const typ $ numeral
    val numeral =
      if n = 0 then
       Const (@{const_name "zero_class.zero"}, typ)
      else if n > 0 then
        wrap (HOLogic.mk_numeral n)
      else
        Const (@{const_name "uminus"}, dummyT) $ wrap (HOLogic.mk_numeral (~n))
  in
    Term_XML.Encode.term numeral
  end;
*}ML{*
Term_XML.Encode.term : term Term_XML.Encode.T;
XML.Encode.string    : string XML.Encode.T;
XML.Decode.string    : string XML.Decode.T
*}text{*
Type error in function application. 
Function: exec : ('a -> Encode.body) -> 'a -> unit 
Argument: (Operations.test_fun) : 'a -> string -> Encode.body
Reason: Can't unify XML.tree list to string -> Encode.body (Incompatible types)
*}ML{*
*}ML{*
*}
end
